import json
from pathlib import Path

import click

from joblib import Parallel, delayed

from ..common.context import slice_context
from ..common.query import match


@click.command()
@click.option("-p", type=int, default=1, help="Number of process.")
@click.option("--name", type=str, help="File name pattern to find dicom file.")
@click.option("--query", type=str, help="Query to find dicom file.")
@click.argument("root", nargs=1)
def find(p, name, query, root):
    name = name or "*"
    query = query or "{}"
    query_obj = json.loads(query)
    if query_obj is None:
        click.echo("{} is not supported query.".format(query))
        return

    def _find_task(path):
        with slice_context(str(path), None) as slice_proxy:
            if match(query_obj, slice_proxy):
                click.echo(path)

    Parallel(n_jobs=p)(
        [delayed(_find_task)(p) for p in Path(root).glob("**/{}".format(name))]
    )



def _or(obj, value):
    result = False
    for query in value:
        result |= _match(obj, query)
    return result


def _and(obj, value):
    result = True
    for query in value:
        result &= _match(obj, query)
    return result


def _eq(obj, value):
    return obj == value


def _ne(obj, value):
    return not _eq(obj, value)


def _in(obj, value):
    return obj in value


def _gt(obj, value):
    return value < obj


def _gte(obj, value):
    return _gt(obj, value) or _eq(obj, value)


def _lt(obj, value):
    return not _gte(obj, value)


def _lte(obj, value):
    return not _gt(obj, value)


def _get_op(key):
    def _eq_with_key(obj, value):
        obj_attr = getattr(obj, key) if hasattr(obj, key) else obj[key]
        return _eq(obj_attr, value)

    return {
        "$eq": _eq,
        "$ne": _ne,
        "$in": _in,
        "$gt": _gt,
        "$gte": _gte,
        "$lt": _lt,
        "$lte": _lte,
    }.get(key, _eq_with_key)


def _match(obj, query):
    """
    >>> _match({}, {})
    True
    >>> _match({'a': 100}, {'a': 100})
    True
    >>> _match({'a': 100}, {'a': 200})
    False
    >>> _match({'a': {'b': {'c': 'd'}}}, {'a': {'b' : {'c': 'd'}}})
    True
    >>> _match({'a': 'b'}, {'a': {'$in' : ['c', 'd']}})
    False
    >>> _match({'a': 'b'}, {'a': {'$in' : ['c', 'b']}})
    True
    >>> _match({'a': 100}, {'a': {'$gt' : 50, '$lt': 200}})
    True
    >>> _match({'a': 100}, {'a': {'$gt' : 50, '$lte': 100}})
    True
    >>> _match({'a': 100}, {'a': {'$ne' : 50}})
    True
    """
    result = True
    for key, value in query.items():
        if isinstance(value, dict):
            ret = _match(obj[key], value)
        else:
            ret = _get_op(key)(obj, value)
        result &= ret
    return result


def match(query, slice_proxy):
    try:
        result = _match(slice_proxy, query)
    except Exception:
        return False
    return result



import json

from pydicom import datadict, dcmread
from pydicom.multival import MultiValue
from pydicom.uid import UID

from ..common.dicom import create_dataset_of


def dump_value(value):
    def repr_value(value):
        import numbers

        if isinstance(value, numbers.Real) or isinstance(value, bool):
            return value

        if not isinstance(value, str):
            value = repr(value)
        return value.rstrip("\x00\u0000")

    def is_array_value(value):
        """
        Return a str representation of the element's `value`.
        The following codes from pydicom repository.
        """
        byte_VRs = [
            "OB",
            "OW",
            "OW/OB",
            "OW or OB",
            "OB or OW",
            "US or SS or OW",
            "US or SS",
        ]
        return value.VR in byte_VRs and len(value.value) > value.maxBytesToDisplay

    def is_multiple_values(value):
        tag = value.tag
        if not tag.is_private:
            vvm = datadict.get_entry(tag)[1].split("-")
            return len(vvm) == 2 or 1 < int(vvm[0])
        elif hasattr(tag, "private_creator"):
            vvm = datadict.get_private_entry(tag, tag.private_creator)[1].split("-")
            return len(vvm) == 2 or 1 < int(vvm[0])
        else:
            return 1

    if is_array_value(value):
        repVal = "Array of %d bytes" % len(value.value)
    elif is_multiple_values(value):
        repVal = (
            [repr_value(v) for v in value]
            if 1 < value.VM
            else [repr_value(value.value)]
        )
    elif isinstance(value.value, UID):
        repVal = value.value.name
    else:
        repVal = repr_value(value.value)
    return repVal


def dump(proxy):
    def _convert_to_dict(ds):
        converted = {}
        for key in sorted(ds.keys()):
            element = ds[key]
            dumped_key = str(key).strip("()")
            dumped_value = (
                [_convert_to_dict(dataset) for dataset in element.value]
                if element.VR == "SQ"
                else {
                    "vr": element.VR,
                    "description": element.description(),
                    "value": dump_value(element),
                }
            )
            converted[dumped_key] = dumped_value
        return converted

    return json.dumps(_convert_to_dict(create_dataset_of(proxy)))



from enum import Enum
from pathlib import Path

import click

from ..common.window import assign as assign_window
from ..common.context import slice_context
from .json import dump as dump_json
from .pretty import dump as dump_pretty
from .terminal import dump_braille, dump_halfblock, dump_iterm2


class Format(Enum):
    PRETTY = "pretty"
    JSON = "json"
    BRAILLE = "braille"
    HALFBLOCK = "halfblock"
    ITERM2 = "iterm2"


def _get_dump_func(format: Format):
    return {
        Format.PRETTY: dump_pretty,
        Format.JSON: dump_json,
        Format.BRAILLE: dump_braille,
        Format.HALFBLOCK: dump_halfblock,
        Format.ITERM2: dump_iterm2,
    }[format]


@click.command()
@click.option(
    "--format",
    type=click.Choice(Format.__members__),
    callback=lambda c, p, v: getattr(Format, v) if v else None,
    default=Format.PRETTY,
)
@click.option("--window", type=str, help="Window configuration")
@click.argument("src", type=click.Path(exists=True), nargs=1)
def dump(format: Format, window: str, src: Path):
    with slice_context(str(src), None) as slice_proxy:
        assign_window(slice_proxy, window)
        dump_func = _get_dump_func(format)
        click.echo(dump_func(slice_proxy))



import click

from .convert import convert
from .dump import dump
from .find import find

cmd = click.Group()
cmd = click.version_option()(cmd)

cmd.add_command(dump)
cmd.add_command(find)
cmd.add_command(convert)


def main():
    cmd()



import numpy as np

from teimpy import Mode, get_drawer

from ..common.window import apply as apply_window


def _dump(mode, proxy):
    drawer = get_drawer(mode)
    return drawer.draw(apply_window(proxy))


def dump_braille(proxy):
    return _dump(Mode.BRAILLE, proxy)


def dump_halfblock(proxy):
    return _dump(Mode.HALF_BLOCK, proxy)


def dump_iterm2(proxy):
    return _dump(Mode.ITERM2_INLINE_IMAGE, proxy)



from abc import ABC, abstractproperty

import numpy as np

from .format import SliceFormat


class proxy_property:
    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __get__(self, inst, objtype=None):
        if inst is None:
            return self

        val = self.wrapped(inst)
        setattr(inst, self.wrapped.__name__, val)
        return val


class SliceProxy(ABC):
    def __init__(self, params=None):
        if params is not None:
            self.__dict__.update(params)

    @property
    def width(self):
        return self.pixels.shape[1]

    @property
    def height(self):
        return self.pixels.shape[0]

    @abstractproperty
    def pixels(self):
        raise NotImplementedError("pixels must be implemented.")

    @abstractproperty
    def spacing(self):
        raise NotImplementedError("spacing must be implemented.")

    @abstractproperty
    def origin(self):
        raise NotImplementedError("origin must be implemented.")

    @abstractproperty
    def orientation(self):
        raise NotImplementedError("orientation must be implemented.")

    @abstractproperty
    def ref(self):
        raise NotImplementedError("ref must be implemented.")


class ImageSliceProxy(SliceProxy):
    def __init__(self, pixel_array, format=None, params=None):
        super().__init__(params)
        self._pixel_array = pixel_array
        self._format = SliceFormat.UNKNOWN if format is None else format
        self._params = params

    @proxy_property
    def pixels(self):
        return self._pixel_array.copy()

    @proxy_property
    def spacing(self):
        return np.array([np.nan, np.nan])

    @proxy_property
    def origin(self):
        return np.array([np.nan, np.nan, np.nan])

    @proxy_property
    def orientation(self):
        return np.array([[np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]])

    @property
    def format(self):
        return self._format

    @property
    def ref(self):
        return (self._pixel_array, self._format, self._params)


class DicomSliceProxy(SliceProxy):

    _PROXY_KEY_MAP = {
        "PixelSpacing": "spacing",
        "ImagePositionPatient": "origin",
        "ImageOrientationPatient": "orientation",
    }

    def __init__(self, dcm, params=None):
        super().__init__(params)
        self._dcm = dcm
        self._params = params

    @proxy_property
    def pixels(self):
        if hasattr(self._dcm, "pixel_array"):
            return self._dcm.pixel_array.copy()
        return np.array([], shape=(0, 0))

    @proxy_property
    def spacing(self):
        if hasattr(self._dcm, "PixelSpacing"):
            return np.array(self._dcm.PixelSpacing)
        if hasattr(self._dcm, "ImagerPixelSpacing"):
            return self._dcm.ImagerPixelSpacing
        return np.array([np.nan, np.nan])

    @proxy_property
    def origin(self):
        if hasattr(self._dcm, "ImagePositionPatient"):
            return np.array(self._dcm.ImagePositionPatient)
        return np.array([np.nan, np.nan, np.nan])

    @proxy_property
    def orientation(self):
        if hasattr(self._dcm, "ImageOrientationPatient"):
            return np.array(self._dcm.ImageOrientationPatient).reshape(2, 3)
        return np.array([[np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]])

    @property
    def ref(self):
        return (self._dcm, SliceFormat.DICOM, self._params)

    def __getattr__(self, key):
        proxy_key = self._PROXY_KEY_MAP.get(key, key)
        if proxy_key != key:
            return getattr(self, proxy_key)

        if hasattr(self._dcm, key):
            value = self._dcm.get(key)
            self.__dict__[key] = value
            return value
        raise AttributeError(f"'DicomSliceProxy' object has no attribute '{key}'")

    def __setattr__(self, key, value):
        key = self._PROXY_KEY_MAP.get(key, key)
        self.__dict__[key] = value



import re
import sys
from typing import Optional, IO
from contextlib import contextmanager

from . import io
from .format import SliceFormat
from .proxy import DicomSliceProxy, ImageSliceProxy


def _parse_dst_path(path: str):
    """
    Parse given imagemagick like path to get format and destination path.

    Parameters
    ----------
    path : str
        Imagemagick linke path such as format:path.

    >>> _parse_dst_path("png:output.dcm")
    (<SliceFormat.PNG: 'PNG'>, 'output.dcm')
    >>> _parse_dst_path("output.dcm")
    (<SliceFormat.DICOM: 'DICOM'>, 'output.dcm')
    >>> _parse_dst_path("output")
    (<SliceFormat.UNKNOWN: 'UNKNOWN'>, 'output')
    >>> _parse_dst_path("foo:output.dcm")
    Traceback (most recent call last):
      ...
    dcmagick.common.exception.NotSupportFormatError: Couldn't get format from foo.
    """
    m = re.match(r"^\s*([^:]*):(.*)$", path)
    if m is not None:
        return SliceFormat.of(m.group(1)), m.group(2)
    m = re.match(r"^\s*(.*)\.(.*)$", path)
    if m is not None:
        return SliceFormat.of(m.group(2)), path
    return SliceFormat.UNKNOWN, path


@contextmanager
def slice_context(src_path: str, dst_path: Optional[str], params: dict = None):
    """
    Slice-wise processing context

    Parameters
    ----------
    src_path : str
        String path to source file.
    dst_path : str
        Imagemagick like path to dstination file.
    params : dict
        Parameters to any features in context. This argument may takes the follwing keys.
        Values to each keys should be dict and contain any value.
            * input  : Parameters for input process such as file reading.
            * proxy  : Parameters for proxy object. This is used for overriding its attributes.
            * output : Parameters for output process such as file writing.
    """

    params = {} if params is None else params
    src_fo = sys.stdin if src_path == "-" else open(src_path, "rb")
    input_params = params.get("input", {})
    proxy_params = params.get("proxy", {})
    output_params = params.get("output", {})
    try:
        slice_proxy = io.read(
            src_fo, input_params=input_params, proxy_params=proxy_params
        )
        yield slice_proxy

        if dst_path is not None:
            dst_format, real_dst_path = _parse_dst_path(dst_path)
            dst_fo = (
                sys.stdout.buffer if real_dst_path == "-" else open(real_dst_path, "wb")
            )
            try:
                io.write(dst_fo, slice_proxy, dst_format, params=output_params)
            finally:
                dst_fo.flush()
                if real_dst_path != "-":
                    dst_fo.close()
    finally:
        if src_path != "-":
            src_fo.close()  # type: ignore



import numpy as np


def normalize_cast(arr, maxv=None, minv=None, dtype=None):
    maxv = np.max(arr) if maxv is None else maxv
    minv = np.min(arr) if minv is None else minv
    dtype = arr.dtype if dtype is None else dtype

    result = (arr - minv) / (maxv - minv)
    if issubclass(dtype.type, np.integer):
        result = np.round(np.iinfo(dtype).max * result)

    return result.astype(dtype)



import click
import numpy as np

from ..common.context import slice_context
from ..common.window import assign as assign_window


@click.command()
@click.option("--window", type=str, help="Window configuration")
@click.argument("src", type=click.Path(exists=True), nargs=1)
@click.argument("dst", nargs=1)
def convert(window, src, dst):
    with slice_context(str(src), dst) as slice_proxy:
        assign_window(slice_proxy, window)



from collections import namedtuple
from copy import deepcopy
from functools import singledispatch

import numpy as np
from pydicom.dataset import Dataset

from .. import __version__ as version
from . import uid
from .proxy import DicomSliceProxy, ImageSliceProxy

PixelInfo = namedtuple(
    "PixelInfo",
    [
        "samples_per_pixel",
        "photometric_interpretation",
        "bits_allocated",
        "pixel_representation",
    ],
)


@singledispatch
def create_dataset_of(slice_proxy):
    import pdb

    pdb.set_trace()
    raise ValueError("Unknown slice proxy is given.")


@create_dataset_of.register(DicomSliceProxy)
def _d(slice_proxy):
    ds = deepcopy(slice_proxy.ref[0])
    for k, v in slice_proxy.__dict__.items():
        if k.startswith("_"):
            continue
        elif k == "pixels":
            _assign_pixel_data(ds, v)
        elif k == "spacing":
            ds.PixelSpacing = list(v)
        elif k == "origin":
            ds.ImagePositionPatient = list(v)
        elif k == "orientation":
            ds.ImageOrientationPatient = list(v.reshape(-1))
        else:
            setattr(ds, k, v)
    return ds


@create_dataset_of.register(ImageSliceProxy)
def _i(slice_proxy):
    sop_inst_uid = uid.generate_uid()

    file_meta = Dataset()
    file_meta.MediaStorageSOPClassUID = (
        uid.storage_sopclass_uids.SecondaryCaptureImageStorage
    )
    file_meta.MediaStorageSOPInstanceUID = sop_inst_uid
    file_meta.ImplementationClassUID = uid.DCMAGICK_IMPLEMENTATION_UID
    file_meta.ImplementationVersionName = version
    file_meta.FileMetaInformationGroupLength = 0
    file_meta.FileMetaInformationVersion = b"\x00\x01"
    file_meta.TransferSyntaxUID = uid.ImplicitVRLittleEndian

    ds = Dataset()
    ds.file_meta = file_meta
    ds.preamble = b"\0" * 128
    ds.is_implicit_VR = True
    ds.is_little_endian = True
    ds.ImageType = ["ORIGINAL"]

    for k, v in slice_proxy.__dict__.items():
        if k.startswith("_"):
            continue
        elif k == "pixels":
            _assign_pixel_data(ds, v)
        elif k == "spacing":
            ds.PixelSpacing = list(v)
        elif k == "origin":
            ds.ImagePositionPatient = list(v)
        elif k == "orientation":
            ds.ImageOrientationPatient = list(v.reshape(-1))
        else:
            setattr(ds, k, v)
    return ds


def _assign_pixel_data(ds, pixels):
    """
    Assign pixels and its information to dataset.

    >>> ds = Dataset()
    >>> _assign_pixel_data(ds, np.array([[0]], dtype=np.uint16))
    >>> assert ds.SamplesPerPixel == 1
    >>> assert ds.PhotometricInterpretation == "MONOCHROME1"
    >>> assert ds.BitsAllocated == 16
    >>> assert ds.BitsStored == 16
    >>> assert ds.HighBit == 15
    >>> assert ds.PixelRepresentation == 0
    >>> assert ds.Rows == 1
    >>> assert ds.Columns == 1
    >>> assert ds.PixelData == np.array([[0]], dtype=np.uint16).tobytes()
    """
    if not _is_rgb888(pixels) and not _is_mono(pixels):
        raise ValueError("Given not supported pixel type.")

    ds.PixelData = pixels.tobytes()
    ds.Rows = pixels.shape[0]
    ds.Columns = pixels.shape[1]

    pixel_info = _get_pixel_info(pixels)
    ds.SamplesPerPixel = pixel_info.samples_per_pixel
    ds.PhotometricInterpretation = pixel_info.photometric_interpretation
    ds.BitsAllocated = pixel_info.bits_allocated
    ds.BitsStored = ds.BitsAllocated
    ds.HighBit = ds.BitsAllocated - 1
    ds.PixelRepresentation = pixel_info.pixel_representation


def _get_pixel_info(pixels):
    """
    Get information about pixels.

    >>> _get_pixel_info(np.array([[0]], dtype=np.uint8))
    PixelInfo(samples_per_pixel=1, photometric_interpretation='MONOCHROME1', bits_allocated=8, pixel_representation=0)
    >>> _get_pixel_info(np.array([[[0, 0, 0]]], dtype=np.uint8))
    PixelInfo(samples_per_pixel=3, photometric_interpretation='RGB', bits_allocated=8, pixel_representation=0)
    >>> _get_pixel_info(np.array([[0]], dtype=np.uint16))
    PixelInfo(samples_per_pixel=1, photometric_interpretation='MONOCHROME1', bits_allocated=16, pixel_representation=0)
    >>> _get_pixel_info(np.array([[0]], dtype=np.int16))
    PixelInfo(samples_per_pixel=1, photometric_interpretation='MONOCHROME1', bits_allocated=16, pixel_representation=1)
    """
    samples_per_pixel = 3 if _is_rgb888(pixels) else 1
    photometric_interpretation = "RGB" if _is_rgb888(pixels) else "MONOCHROME1"
    bits_allocated = 8 * pixels.dtype.itemsize
    pixel_representation = 1 if pixels.dtype == np.int16 else 0

    return PixelInfo(
        samples_per_pixel,
        photometric_interpretation,
        bits_allocated,
        pixel_representation,
    )


def _is_rgb888(pixels):
    """
    Predicate function for RGB888 or not.

    >>> _is_rgb888(np.array([[[0, 0, 0]]], dtype=np.uint16))
    False
    >>> _is_rgb888(np.array([[[0, 0, 0]]], dtype=np.uint8))
    True
    >>> _is_rgb888(np.array([[0]]))
    False
    """

    is_rgb = pixels.ndim == 3 and pixels.shape[-1] == 3
    is_unit8 = pixels.dtype == np.uint8
    return is_rgb and is_unit8


def _is_mono(pixels):
    """
    Predicate function for monochrome or not.

    >>> _is_mono(np.array([[[0, 0, 0]]]))
    False
    >>> _is_mono(np.array([[0]]))
    True
    >>> _is_mono(np.array([[[0]]]))
    True
    """

    return pixels.ndim == 2 or (pixels.ndim == 3 and pixels.shape[-1] == 1)



import pydicom


def slice_location_order(dcm):
    return dcm.SliceLocation


def read_dcms(dcm_paths):
    result = {}
    for path in dcm_paths:
        dcm = pydicom.dcmread(path)
        result.setdefault(dcm.StudyInstanceUID, {}).setdefault(
            dcm.SeriesInstanceUID, {}
        )[dcm.SOPInstanceUID] = dcm

    return result



import numpy as np
from PIL import Image
from pydicom import dcmread

from .dicom import create_dataset_of
from .format import SliceFormat
from .proxy import DicomSliceProxy, ImageSliceProxy


def read(fo, input_params: dict = None, proxy_params: dict = None):
    _read_funcs = [_read_image, _read_dcm]

    input_params = {} if input_params is None else input_params
    proxy_params = {} if proxy_params is None else proxy_params
    for f in _read_funcs:
        try:
            return f(fo, input_params, proxy_params)
        except Exception:
            fo.seek(0)
    raise ValueError("Unknown format data is given.")


def _read_image(fo, input_params, src_params):
    img = Image.open(fo, **input_params)
    pixel_array = np.asarray(img)
    try:
        slice_format = SliceFormat(img.format)
    except ValueError:
        slice_format = SliceFormat.UNKNOWN
    return ImageSliceProxy(pixel_array, format=slice_format, params=src_params)


def _read_dcm(fo, input_params, src_params):
    dcmread_params = {
        k: src_params[k]
        for k in ["defer_size", "stop_before_pixels", "force", "specific_tags"]
        if k in src_params
    }
    dcm = dcmread(fo, **dcmread_params)
    return DicomSliceProxy(dcm, params=src_params)


def write(fo, slice_proxy, format=SliceFormat.UNKNOWN, params: dict = None):
    format = slice_proxy.ref[1] if format == SliceFormat.UNKNOWN else format
    params = {} if params is None else params

    if format == SliceFormat.DICOM:
        _write_dcm(fo, slice_proxy, params)
    elif format in [SliceFormat.JPEG, SliceFormat.PNG]:
        _write_image(fo, slice_proxy, format, params)
    else:
        raise ValueError(f"Unknown format({format}) is given.")


def _write_image(fo, slice_proxy, format, params):
    pixels = slice_proxy.pixels
    if pixels.dtype != np.uint8:
        min_value = float(np.min(pixels))
        max_value = float(np.max(pixels))
        pixels = (255 * (pixels - min_value) / (max_value - min_value)).astype(np.uint8)

    img = Image.fromarray(pixels)
    img.save(fo, format=format.value, **params)


def _write_dcm(fo, slice_proxy, params):
    ds = create_dataset_of(slice_proxy)
    ds.save_as(fo, **params)



from enum import Enum

from .exception import NotSupportFormatError


class SliceFormat(Enum):
    DICOM = "DICOM"
    JPEG = "JPEG"
    PNG = "PNG"
    UNKNOWN = "UNKNOWN"

    @classmethod
    def of(cls, format: str):
        """
        Get SliceFormat of given string.

        >>> SliceFormat.of("dcm")
        <SliceFormat.DICOM: 'DICOM'>
        >>> SliceFormat.of("jpg")
        <SliceFormat.JPEG: 'JPEG'>
        >>> SliceFormat.of("jpeg")
        <SliceFormat.JPEG: 'JPEG'>
        >>> SliceFormat.of("png")
        <SliceFormat.PNG: 'PNG'>
        """

        if format in ["dcm"]:
            return SliceFormat.DICOM
        elif format in ["jpg", "jpeg"]:
            return SliceFormat.JPEG
        elif format in ["png"]:
            return SliceFormat.PNG
        raise NotSupportFormatError(f"Couldn't get format from {format}.")



import re
from collections import Sequence

import numpy as np

from .exception import WindowFormatError
from .improc import normalize_cast
from .proxy import DicomSliceProxy, SliceProxy


def _parse_window(window: str, slice_proxy: SliceProxy):
    """
    Parse given window string.

    >>> from .proxy import ImageSliceProxy
    >>> proxy = ImageSliceProxy([[10, 50]])
    >>> _parse_window("100x200", proxy)
    (100.0, 200.0)
    >>> _parse_window("100x200abc", proxy)
    Traceback (most recent call last):
    ...
    dcmagick.common.exception.WindowFormatError: 100x200abc is invalid for --window.
    >>> _parse_window("0", proxy)
    (None, None)
    >>> proxy.WindowCenter = [200, 500]
    >>> proxy.WindowWidth = [300, 800]
    >>> _parse_window("0", proxy)
    (200.0, 300.0)
    >>> _parse_window("1", proxy)
    (500.0, 800.0)
    >>> _parse_window("2", proxy)
    (None, None)
    >>> proxy.WindowCenter = 200
    >>> proxy.WindowWidth = 300
    >>> _parse_window("0", proxy)
    (200.0, 300.0)
    >>> _parse_window("1", proxy)
    (None, None)
    >>> _parse_window(None, proxy)
    (30.0, 20.0)
    """
    if window is None:
        maxv = np.max(slice_proxy.pixels)
        minv = np.min(slice_proxy.pixels)
        wc = (maxv + minv) / 2
        ww = (maxv - minv) / 2
        return wc, ww

    m = re.match(r"^(\d+)$", window)
    if m is not None:
        index = int(m.group(1))
        wc_candidates = getattr(slice_proxy, "WindowCenter", [])
        if not isinstance(wc_candidates, Sequence):
            wc_candidates = [wc_candidates]
        wc = float(wc_candidates[index]) if index < len(wc_candidates) else None

        ww_candidates = getattr(slice_proxy, "WindowWidth", [])
        if not isinstance(ww_candidates, Sequence):
            ww_candidates = [ww_candidates]
        ww = float(ww_candidates[index]) if index < len(ww_candidates) else None
        return wc, ww

    m = re.match(r"^(\d+)x(\d+)$", window)
    if m is not None:
        wc = float(m.group(1))
        ww = float(m.group(2))
        return wc, ww

    raise WindowFormatError(f"{window} is invalid for --window.")


def assign(slice_proxy: SliceProxy, window: str):
    wc, ww = _parse_window(window, slice_proxy)
    if isinstance(slice_proxy, DicomSliceProxy) and (wc is None or ww is None):
        return

    setattr(slice_proxy, "WindowCenter", wc)
    setattr(slice_proxy, "WindowWidth", ww)


def apply(slice_proxy: SliceProxy) -> np.ndarray:
    wc = getattr(slice_proxy, "WindowCenter")
    wc = wc[0] if isinstance(wc, Sequence) else wc
    ww = getattr(slice_proxy, "WindowWidth")
    ww = ww[0] if isinstance(ww, Sequence) else ww

    if wc is None or ww is None:
        return normalize_cast(slice_proxy.pixels, dtype=np.dtype("uint8"))

    half_ww = ww // 2
    minv = wc - half_ww
    maxv = wc + half_ww
    windowed = np.clip(slice_proxy.pixels, minv, maxv)
    return normalize_cast(windowed, maxv=maxv, minv=minv, dtype=np.dtype("uint8"))



from pydicom import _storage_sopclass_uids as storage_sopclass_uids
from pydicom.uid import *
from pydicom.uid import PYDICOM_IMPLEMENTATION_UID

DCMAGICK_IMPLEMENTATION_UID = PYDICOM_IMPLEMENTATION_UID
